<?php

/** 
 * This class is used to classify the query string as Java or Python
 * @author Snigdha Rao Parvatneni snigdha.parvatneni@gmail.com
 * @copyright 2013
 */
class classifier 
{
    /**
     * This function is used to check number of command line arguments
     * 
     * If the command line arguments passed to this function is less 
     * than 6 then this function will display the error message
     * @param integer $argc number of command line arguments passed
     * @param array $argv actual of array of values passed via command line
     * @return boolean $flag boolean value returned
     */
    public function isOkToProceed($argc, $argv) 
    {
        if($argc != 6)
        {
            $flag = FALSE;
            echo "Invalid arguments \n";
        } else {
            $flag = TRUE;
        }
        return $flag;
    }

    /**
     * This function is used to read the contents of text file
     * 
     * Read the contents of text file in an array where end of the document
     * is marked by \n\n.
     * @param string $filename actural name of the file
     * @return string $contents actual content of the file
     */
    public function readContents($filename) 
    {
        $contents = explode("\n\n", file_get_contents($filename));
        return $contents;
    }

    /**
     * This function is to split the data into trigrams 
     * 
     * Chunk the actual content of the file into trigrams and record the
     * document number where it appears
     * @param string $contents actual contents of the files
     * @return array $sortedArray keys are the trigrams and values are the
     *         document numbers 
     */
    public function countWords($contents) 
    {
        $sortedArray = array();
        $docNumber = -1;
        foreach($contents as $content) {
            $docNumber = $docNumber + 1;
            $details = explode("\n", $content);
            foreach($details as $detail) {
                if(strlen($detail) > 3) {
                    $longDetails = $detail;
                    $word =  $longDetails;
                    for ($i = 0; $i < ((strlen($longDetails)) - 2); $i++) {
                        $trigrams = substr($word, $i, 3);
                        if (array_key_exists($trigrams, $sortedArray)) {
                            $sortedArray = classifier::addNewTrigram(
                                           $sortedArray, $trigrams, $docNumber);
                        } else {
                            $sortedArray = classifier::addExistingTrigram(
                                           $sortedArray, $trigrams, $docNumber);
                        }
                    }
                } else {
                    $shortDetails = $detail;
                    if (array_key_exists($shortDetails, $sortedArray)) {
                        $sortedArray = classifier::addNewWord($sortedArray,
                                       $shortDetails, $docNumber);
                    } else {
                        $sortedArray = classifier::addExistingWord($sortedArray, 
                                       $shortDetails, $docNumber);
                    }
                }
            }
        }
        return $sortedArray;
    }

    /**
     * This funtion is used to add new trigrams which were not already present
     * 
     * @param array $sortedArray keys are the trigrams and values are the
     *        document numbers
     * @param string $trigrams trigrams from each file passed in command line
     *        arguments
     * @param integer $docNumber finds document number where trigrams appear
     * @return array $sortedArray keys are the trigrams and values are the
     *         document numbers
     */
    function addNewTrigram($sortedArray, $trigrams, $docNumber)
    {
        $documentCount = $sortedArray[$trigrams];
        if(!in_array($docNumber, $documentCount)) {
            $documentCount[] = $docNumber;
            $sortedArray[$trigrams] =  $documentCount;
        }
        return $sortedArray;
    }

    /**
     * This funtion is used to add new document numbers for existing trigrams
     * 
     * @param array $sortedArray keys are the trigrams and values are the
     *        document numbers
     * @param string $trigrams trigrams from each file passed in command line
     *        arguments
     * @param integer $docNumber finds document number where trigrams appear
     * @return array $sortedArray keys are the trigrams and values are the
     *         document numbers
     */
    function addExistingTrigram($sortedArray, $trigrams, $docNumber)
    {
        $docCount = array();
        $docCount[0] = $docNumber;
        $sortedArray[$trigrams] =  $docCount;
        return $sortedArray;
    }

    /**
     * This funtion is used to add new worls length less than two which were
     * not already present
     * 
     * @param array $sortedArray keys are the trigrams and values are the
     *        document numbers
     * @param string $trigrams trigrams from each file passed in command line
     *        arguments
     * @param integer $docNumber finds document number where trigrams appear
     * @return array $sortedArray keys are the trigrams and values are the
     *         document numbers
     */
    function addNewWord($sortedArray, $shortDetails, $docNumber)
    {
        $documentID = $sortedArray[$shortDetails];
        if(!in_array($docNumber, $documentID)) {
            $documentID[] = $docNumber;
            $sortedArray[$shortDetails] =  $documentID;
        }
        return $sortedArray;
    }

    /**
     * This funtion is used to add new document numbers to the existing words
     * which were already present
     * 
     * @param array $sortedArray keys are the trigrams and values are the
     *        document numbers
     * @param string $trigrams trigrams from each file passed in command line
     *        arguments
     * @param integer $docNumber finds document number where trigrams appear
     * @return array $sortedArray keys are the trigrams and values are the
     *         document numbers
     */
    function addExistingWord($sortedArray, $shortDetails, $docNumber)
    {
        $docId = array();
        $docId[0] = $docNumber;
        $sortedArray[$shortDetails] =  $docId;
        return $sortedArray;
    }

    /**
     * This function splits the query into trigrams
     * 
     * @param string $query query passed in command line argument
     * @return array $qrTrigrams contains trigrams as array values from query
     */
    public function splitQuery($query)
    {
        $qrTrigrams = array();
        if(strlen($query) > 3) {
            for ($i = 0; $i < ((strlen($query))-2); $i++) {
                $trigrams = substr($query, $i, 3);
                    if(!in_array($trigrams, $qrTrigrams)) {
                        $qrTrigrams[] = $trigrams;
                    }
            }
         } else {
             if(!in_array($query,$qrTrigrams)) {
                 $qrTrigrams[] = $query;
             }
         }
         return $qrTrigrams;
    }

    /**
     * This method is used to calculate probabilities of hypothesis i.e
     * probabilities of Java and Python
     * 
     * @param string $language name of two programming languages compared
     * @return BCMath number $resultProb hypothesis probability
     */
    public function hypothesisProbability($language)
    {
        $countJava = '177000000.00';
        $countPython = '42000000.00';
		//constants
		define("HUNDERD", "100");
		define("HUNDERD_ONE", "101");
		
        $countTotal = bcadd($countJava, $countPython, HUNDERD);
        $javaProb = bcdiv ($countJava,$countTotal, HUNDERD_ONE);
        $pythonProb = bcdiv ($countPython,$countTotal, HUNDERD_ONE);
        if($language == 'Java') {
            $resultProb = roundbc($javaProb, HUNDERD);
        }
        if($language == 'Python') {
            $resultProb = roundbc($pythonProb, HUNDERD);
        }
        return $resultProb;
    }

    /**
     * This method is to calculate the initial probability of terms appear in
     * Java training set
     * 
     * @param array $qrTrigrams contains trigrams as array values from query
     * @param array $sortedArray keys are the trigrams and values are the
     *        document numbers
     * @param string $contents actual content of the file
     * @return array $sortedProbArray contains keys as trigrams and values as
     *         probabilities of trigrams
    */
    public function initialProbability($qrTrigrams,$sortedArray,$contents)
    {
        $keyArray = array_keys($sortedArray);
        $totalDocCount = count($contents);
        $sortedProbArray = array();
        $i = 0;
		//constants
		define("HUNDERD", "100");
		define("HUNDERD_ONE", "101");
        foreach ($sortedArray as $tempArray) {
            $tempKey = $keyArray[$i];
            $tempCount = count($tempArray);
            $iniProb =  bcdiv($tempCount, $totalDocCount, HUNDERD_ONE);
            $iniProb = roundbc($iniProb, HUNDERD);
            $sortedProbArray[$tempKey] = $iniProb;
            $i = $i + 1;
        }
        return $sortedProbArray;
    }

    /**
     * This function calculates the probability of unknown trigrams
     * 
     * @param array $sortedArray keys are the trigrams and values are the
              document numbers
     * @param array $sortedRandomArray keys are the trigrams and values are
     *        the document numbers
     * @return array $unknownProb contains probability for unknown trigrams
     */
    public function unknownProbability($sortedArray,$sortedRandomArray)
    {
    	//constants
		define("HUNDERD", "100");
		define("HUNDERD_ONE", "101");
        $KeyArray = array_keys($sortedArray);
        $randomKeyArray = array_keys($sortedRandomArray);
        $trigramCount = count($randomKeyArray);
        $unknownCount = 0;
        for($i = 0; $i< count($randomKeyArray); $i++) {
            if(!(in_array($randomKeyArray[$i],$KeyArray))) {
                $unknownCount = $unknownCount+1;
            }
        }
        $unknownProb = bcdiv($unknownCount, $trigramCount, HUNDERD_ONE);
        $unknownProb = roundbc($unknownProb, HUNDERD);
        return $unknownProb;
    }

    /**
     * This method is used to calculate smoothening factor
     * 
     * @param BCMath Number $unknownProb it is probability of unknown trigram
     * @return BCMath Number $smootheningFactor it is smoothening factor for
     *         known trigrams
     */
    public function smootheningFactor($unknownProb)
    {
    	//constants
		define("HUNDERD", "100");
		define("HUNDERD_ONE", "101");
        $smootheningFactor = bcsub('1', $unknownProb, HUNDERD_ONE);
        $smootheningFactor = roundbc($smootheningFactor, HUNDERD);
        return $smootheningFactor;
    }

    /**
     * This function is used to smoothened the initial probability of trigrams
     * of training set
     * 
     * @param array $initialProb initial probabilities of trigrams
     * @param BCMath Number $smootheningFactor it is smoothening factor for
     *        known trigrams
     */
    public function smoothnedProbability($initialProb, $smootheningFactor) 
    {
    	//constants
		define("HUNDERD", "100");
		define("HUNDERD_ONE", "101");
        $keyArray = array_keys($initialProb);
        $smoothnedProbArray = array();
        for($i = 0; $i<count($initialProb); $i++) {
            $tempKey = $keyArray[$i];
            $tempSmoothProb = bcmul($initialProb[$tempKey], $smootheningFactor,
                                    HUNDERD_ONE);
            $tempSmoothProb = roundbc($tempSmoothProb, HUNDERD);
            $smoothnedProbArray[$tempKey] = $tempSmoothProb;
        }
        return $smoothnedProbArray;
    }

    /**
     * This function will form the array of trigrams of query present in java
       training set
     * 
     * @param array $javaProb it contains smoothned probabilities for known Java
     *        trigrams
     * @param array $query it contains trigrams from query
     * @return array $javaArray contains all the trigrams apperaed in
               Java training set
     */
    public function javaSearch($javaProb, $query) 
    {
        $tempJavaArray = array_keys($javaProb);
        $javaCount = count($tempJavaArray);
        $javaArray = array();
        for($i = 0; $i < $javaCount; $i++) {
            if(in_array($tempJavaArray[$i], $query) &&
              (!in_array($tempJavaArray[$i], $javaArray))) {
                $javaKey = $tempJavaArray[$i];
                $javaArray[$javaKey] = $javaProb[$javaKey];
            }
        }
        return $javaArray;
    }

    /**
     * This function will form the array of trigrams of query present in python
     * training set
     * 
     * @param array $pythonProb it contains smoothned probabilities for known
     *        Python trigrams
     * @param array $query it contains trigrams from query
     * @return array $pythonArray contains all the trigrams apperaed in Python
     *         training set
     */
    public function pythonSearch($pythonProb, $query) 
    {
        $tempPythonArray = array_keys($pythonProb);
        $pythonCount = count($tempPythonArray);
        $pythonArray = array();
        for($i= 0; $i < $pythonCount; $i++) {
            if(in_array($tempPythonArray[$i], $query) &&
              (!in_array($tempPythonArray[$i], $pythonArray))) {
                $pythonKey = $tempPythonArray[$i];
                $pythonArray[$pythonKey] = $pythonProb[$pythonKey];
            }
        }
        return $pythonArray;
    }

    /**
     * This function will form the array of trigrams of query not present in
     * java training set
     * 
     * @param array $javaProb it contains smoothned probabilities for known Java
     *        trigrams
     * @param array $query it contains trigrams from query
     * @return array $unknownArray contains all the unknown trigrams not present
     *         in Java training set
     */
    public function javaUnknownSearch($javaProb, $query) 
    {
        $countQuery = count($query);
        $tempJavaKeys = array_keys($javaProb);	
        $unknownArray = array();
        for($i = 0;$i < $countQuery; $i++){
            if(!in_array($query[$i], $tempJavaKeys)) {
                $unknownKey = $query[$i];
                $unknownArray[] = $unknownKey;
            }
        }
        return $unknownArray;
    }

    /**This function will form the array of trigrams of query not present in 
       python training set
     * 
     * @param array $pythonProb it contains smoothned probabilities for known
     *        Python trigrams
     * @param array $query it contains trigrams from query
     * @return array $pythonArray contains all the unknown trigrams not present
     *         in Python training set
     */
    public function pythonUnknownSearch($pythonProb, $query) 
    {
        $countQuery = count($query);
        $tempPythonKeys = array_keys($pythonProb);
        $unknownArray = array();
        for($i = 0; $i < $countQuery; $i++){
            if(!in_array($query[$i], $tempPythonKeys)) {
                $unknownKey = $query[$i];
                $unknownArray[] = $unknownKey;
            }
        }
        return $unknownArray;
    }

    /**This method calculates the final result for java
     * 
     * @param array $javaArray it contains known trigrams for Java
     * @param array $javaUnknownArray it contains unknown trigrams for Java
     * @param BCMath Number $javaLambda it is the probability of the unknown 
     *        trigrams in Java
     * @param BCMath Number $javaHypProb it is probability of Java hypothesis
     * @return BCMath Number $javaResult it is the final probability of the
     *         query to be from Java
     */
    public function calculateJavaProbability($javaArray, $javaUnknownArray,
                                             $javaLambda, $javaHypProb)
    {
    	//constants
		define("HUNDERD", "100");
		define("HUNDERD_ONE", "101");
        $tempJavaSize = count($javaArray);
        $tempJavaUnknownSize = count($javaUnknownArray);
        $tempJavaKey = array_keys($javaArray);
        $javaResult = 1;
        $unknownResult = 1;
        $ALPHA = 1;
        for($i = 0; $i < $tempJavaSize; $i++) {
            $javaKey = $tempJavaKey[$i];
            $javaIntermediate = $javaArray[$javaKey];
            $javaResult = bcmul($javaIntermediate, $javaResult, HUNDERD_ONE);
        }
        for($j = 0; $j<$tempJavaUnknownSize; $j++) {
            $unknownIntermediate = $javaLambda;
            $unknownResult = bcmul($unknownIntermediate, $unknownResult, 
                                   HUNDERD_ONE);
        }
        $javaResult = bcmul($javaResult, $unknownResult, HUNDERD_ONE);
        $javaResult = bcmul($javaResult, $ALPHA, HUNDERD_ONE);
        $javaResult = bcmul($javaResult, $javaHypProb, HUNDERD_ONE);
        $javaResult = roundbc($javaResult, HUNDERD);
        return $javaResult;
    }

    /**This method calculates the final result for python
     * @param array $pythonArray it contains known trigrams for Python
     * @param array $pythonUnknownArray it contains unknown trigrams for Python
     * @param BCMath Number $pythonLambda it is the probability of the unknown
     *        trigrams in Python
     * @param $pythonHypProb it is probability of Python hypothesis
     * @return BCMath Number $pythonResult it is the final probability of the
     *         query to be from Python
     */
    public function calculatePythonProbability($pythonArray,
                                               $pythonUnknownArray,
                                               $pythonLambda, $pythonHypProb)
    {
    	//constants
		define("HUNDERD", "100");
		define("HUNDERD_ONE", "101");
        $tempPythonSize = count($pythonArray);
        $tempPythonUnknownSize = count($pythonUnknownArray);
        $tempPythonKey = array_keys($pythonArray);
        $pythonResult = 1;
        $unknownResult = 1;
        $ALPHA = 1;
        for($i = 0; $i < $tempPythonSize; $i++) {
            $pythonKey = $tempPythonKey[$i];
            $pythonIntermediate = $pythonArray[$pythonKey];
            $pythonResult = bcmul($pythonIntermediate, $pythonResult,
                                  HUNDERD_ONE);
        }
        for($j = 0; $j < $tempPythonUnknownSize; $j++) {
            $unknownIntermediate = $pythonLambda;
            $unknownResult = bcmul($unknownIntermediate, $unknownResult,
                                   HUNDERD_ONE);
        }
        $pythonResult = bcmul($pythonResult, $unknownResult, HUNDERD_ONE);
        $pythonResult = bcmul($pythonResult, $ALPHA, HUNDERD_ONE);
        $pythonResult = bcmul($pythonResult, $pythonHypProb, HUNDERD_ONE);
        $pythonResult = roundbc($pythonResult, HUNDERD);
        return $pythonResult;
    }

    /**
     * This method will compare final results calculated for java and python and
     * given the query language
     * 
     * @param BCMath Number $javaResult it the final probability for the query
     *        to be from Java
     * @param BCMath Number $pythonResult it the final probability for the query
     *        to be from Python
     * @return string $result it is the final result to be displayed after
     *         comparision
    */
    public function classifyResult($javaResult, $pythonResult) 
    {
    	//constants
		define("HUNDERD", "100");
        $cmpResult = bccomp($javaResult, $pythonResult, HUNDERD);
        if($cmpResult == 1) {
            $result = 'Result--->Query entered belongs to Java programming ' .
                      'language';
        } else if($cmpResult == -1) {
            $result = 'Result--->Query entered belongs to Python programming ' .
                       'language';
        }
        else if($cmpResult == 0) {
            $result = 'Result--->Query entered belongs to Both programming ' .
                      'languages';
        }
        else {
            $result = 'Result--->Query entered neither belongs to Java ' .
                      'programming language nor belongs to Python programming '
                       . 'language';
        }
        return $result;
    }
}
//end of class

/**
 * This method is taken from the internet to round half-up the numbers obtained
 * from bigmath methods
 * 
 * @param BCMath Number $x it is the number to be rounder off
 * @param integer $p it is the precission needed for round off
 * @return BCMath Number $data it is the final rounded off value
 */
function roundbc($x, $p) 
{
    $x = trim($x);
    $data = explode(".", $x);
    if(substr($data[1], $p, 1) >= "5") {
        $i = 0;
        $addString = "5";
        while($i < $p) {
            $addString = "0" . $addString;
            $i++;
        }
        $addString = "." . $addString;
        $sum = bcadd($data[0] . "." . $data   [1],$addString,$p + 1);
        $sumData = explode(".",$sum);
        return $sumData[0] . "." . substr($sumData[1], 0, $p);
    }
    else {
        return $data[0] . "." . substr($data[1], 0, $p);
    }
}

//Creating object of class
$classifier = new classifier();
//Check the number of arguments
$proceedFlag = $classifier->isOkToProceed($argc, $argv);
//if number of arguments are sufficient then only do the other steps
if ($proceedFlag == 1) {
    //reading the contents of files
    $javaContent = $classifier->readContents($argv[1]);
    $javaRandomContent = $classifier->readContents($argv[2]);
    $pythonContent = $classifier->readContents($argv[3]);
    $pythonRandomContent = $classifier->readContents($argv[4]);
    //read and split of query
    $queryContent = $classifier->splitQuery($argv[5]);
    //split the contents of file into trigrams
    $javaArray = $classifier->countWords($javaContent);
    $javaRandomArray = $classifier->countWords($javaRandomContent);
    $pythonArray = $classifier->countWords($pythonContent);
    $pythonRandomArray = $classifier->countWords($pythonRandomContent);
    //calculating hypothesis of probabilities i.e. java and python
    $javaProb = $classifier->hypothesisProbability('Java');
    $pythonProb = $classifier->hypothesisProbability('Python');
    //calculating initial probabilities 
    $javaInitialProb = $classifier->initialProbability($queryContent,
                                                       $javaArray,
                                                       $javaContent);
    $pythonInitialProb = $classifier->initialProbability($queryContent,
                                                         $pythonArray,
                                                         $pythonContent);
    //calculating probabilities of unknown trigrams of java and python
    $javaUnknown = $classifier->unknownProbability($javaArray,
                                                   $javaRandomArray);
    $pythonUnknown = $classifier->unknownProbability($pythonArray,
                                                     $pythonRandomArray);
    //calculating smoothening factor
    $javaSmootheningFactor = $classifier->smootheningFactor($javaUnknown);
    $pythonSmootheningFactor = $classifier->smootheningFactor($pythonUnknown);
    //calculating smoothened probabilities 
    $javaSmoothProb = $classifier->smoothnedProbability($javaInitialProb,
                                                        $javaSmootheningFactor);
    $pythonSmoothProb = $classifier->smoothnedProbability($pythonInitialProb,
                                                      $pythonSmootheningFactor);
    //finding query trigrams in java, python
    $javaSearch = $classifier->javaSearch($javaSmoothProb,$queryContent);
    $pythonSearch = $classifier->pythonSearch($pythonSmoothProb,$queryContent);
    $javaUnknownSearch = $classifier->javaUnknownSearch($javaSmoothProb,
                                                        $queryContent);
    $pythonUnknownSearch = $classifier->pythonUnknownSearch($pythonSmoothProb,
                                                            $queryContent);
    //calculating final probabilities
    $javaResult = $classifier->calculateJavaProbability($javaSearch,
                                                        $javaUnknownSearch,
                                                        $javaUnknown,$javaProb);
    $pythonResult = $classifier->calculatePythonProbability($pythonSearch,
                                                           $pythonUnknownSearch,
                                                           $pythonUnknown,
                                                           $pythonProb);
    $result = $classifier->classifyResult($javaResult,$pythonResult);
    //displayng result
    echo 'Probability (query is from Java)------->'.$javaResult."\n";
    echo 'Probability (query is from Python)----->'.$pythonResult."\n";
    echo $result."\n";
}
?>